#!/usr/bin/env python
# coding: utf-8

import json
import os
import subprocess
import sys
import time
import cgitb
import logging
import logging.handlers
import contextlib
from threading import Thread

import tornado.ioloop
from tornado.stack_context import StackContext
from ws4py.client.tornadoclient import TornadoWebSocketClient

from agenttoserver import auth_response, status_response, log_message, change_installation_status

DEBUG = True
OUTPUT_TO_AGENT_TERMINAL = True


def setup_logger(log_file, rotation=5*1024, count=5):
    """ Настраивает логирование в STDOUT и указанный лог-файл @a log_file.
        Объём ротации задаётся 
    """
    formatter = logging.Formatter('[%(asctime)s][%(levelname)s][%(module)s]: %(message)s', datefmt='%H:%M:%S %d-%m-%Y')
    
    handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=rotation*1024, backupCount=count)
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)
    
    console = logging.StreamHandler()
    console.setFormatter(formatter)
    logging.getLogger().addHandler(console)
    
    logging.getLogger().setLevel(logging.DEBUG)
    

class OutLogThread(Thread):
    """ Поток. который отправляет лог установки на сервер, 
        читает из канала, отправляет через веб сокет
    """
    def __init__(self, agent, pipe_desk, stream_type, *args, **kwargs):
        Thread.__init__(self, *args, **kwargs)
        self.daemon = True
        self._agent = agent
        self._pipe_desk = pipe_desk
        self._stream_type = stream_type
        self._stop = False
        self._buffer = ''
        
    def run(self):
        while True:
            try:
                got = os.read(self._pipe_desk, 255) 
                if got != '':
                    messages = got.split('\n')
                        
                    if self._buffer != '':
                        messages[0] = self._buffer + messages[0]
                        self._buffer = ''
                    
                    if len(got) == 255:
                        self._buffer = messages.pop(-1)
                    
                    for msg in messages:
                        if msg != "" and not msg.startswith("stop: Unknown instance:"):
                            logging.debug(msg)
                            self._agent._send(log_message(self._stream_type, msg))
                else:
                    time.sleep(0.5)
            except:
                break


class AgentWebSocket(TornadoWebSocketClient):
    def __init__(self, address):
        self._address = address
        self._connected = False
        self._agent_name = ''
        self._status = 'Not allowed'
        self._opened_called = False # Для обработки провала handshake, при котором closed вызывается до opened
        self._connection_check_callback = tornado.ioloop.PeriodicCallback(self._check_connection, 5000)
        self._connection_check_callback.start()
        self._connect()
        
    @contextlib.contextmanager
    def _reconnect_on_error(self):
        try:
            yield
        except Exception as e:
            logging.error('Exception: "{0}". Reconnecting in 1 minute.'.format(str(e)))
            self._connected = False

    def _check_connection(self):
        """ Проверка соединения с сервером.
            В случае потери соединения принимается попытка его восстановления.
        """
        self._ping()

    def _connect(self):
        """ Подключение к центральному ДЦ.
        """
        logging.debug("_connect: connected={}".format(self._connected))
        if self._connected:
            return
        
        logging.debug('Connecting to {} ...'.format(self._address))
        TornadoWebSocketClient.__init__(self, self._address)
        with StackContext(self._reconnect_on_error):
            self.connect()
    
    def _send(self, message):
        """ Метод для отправки сообщений. В случае исключений в "родном"
            методе send() инициирует переподключение.
        """
        try:
            self.send(message)
        except Exception as e:
            logging.error('Message could not be sent: {}. Reconnecting.'.format(message))
            print e
            self._connected = False
            self._connect()

    def _ping(self):
        """ Пинг центрального сервера. Нужен для поддержки
            WebSocket-соединения (Keep-Alive).
        """
        self._send({'command': 'ping', 'params': {}})
    
    def opened(self, protocols, extensions):
        self._opened_called = True
        self._connected = True
        
        logging.info('Connected to Junta Server')
        if self._agent_name != '':
            self._send(auth_response(self._agent_name, ''))
            if self._status != 'Not allowed':
                self._send(status_response(self._status))
    
    def closed(self, code, reason):
        """ Связь с центральной базой прервана. 
        """
        logging.warning('Junta Server disconnected')
        if not self._opened_called:
            """ Обработка провалившегося handshake,
                при котором closed вызывается до opened.
            """
            logging.warning("'closed' called before 'opened'")
            return
            
        self._opened_called = False
        
        self._connected = False
        self._connect()
    
    def received_message(self, message):
        message = json.loads(str(message))
        command = message['command']
        
        logging.debug('command received: {0}'.format(command))
        try:
            if command == 'auth_request':
                self._send(auth_response(self._agent_name, ''))
            elif command == 'auth_response':
                logging.debug('Connection {0}'.format(message['response']))
                self._status = 'Online'
                self._send(status_response(self._status))
            elif command == 'status_request':
                self._send(status_response(self._status))
            elif command == 'installation_request':
                time.sleep(0.5)
                self._install(message)
            elif command == 'stop_installation_request':
                pass
            elif command == 'exit_request':
                exit(0)
            else:
                logging.warning('UNKNOWN COMMAND')
        except Exception as e:
            logging.error(cgitb.text(sys.exc_info()))
            
    def _install(self, message):
        if self._status == 'Online' or self._status == 'Error':
            self._status = 'Installation'
            self._send(log_message('main', 'Installation starts.'))
            self._send(change_installation_status('in process'))
            self._send(status_response(self._status))
            try:
                os.chdir('/tmp')
                recipe_file = open('recipe.xml', 'w')
                recipe_file.write(message['recipe'].encode("utf-8"))
                recipe_file.close()
                setup_tools_file = open('setup_tools.py', 'w')
                setup_tools_file.write(message['setup_tools'].encode("utf-8"))
                setup_tools_file.close()
                
                config = message['config']
                file_stdin = open('stdin', 'w')
                logging.debug("Configuration:")
                for c in config:
                    file_stdin.write(c + '\n')
                file_stdin.close()
            
                if OUTPUT_TO_AGENT_TERMINAL:
                    logging.info('Installation starts.')
                os.environ.update(message['config'])
                out_r, out_w = os.pipe()
                err_r, err_w = os.pipe()
                p = subprocess.Popen('python setup_tools.py recipe.xml', shell=True, 
                                     stdout=out_w, stderr=err_w, 
                                     env=os.environ)
        
                err_thread = OutLogThread(self, err_r, 'err')
                err_thread.start()
                out_thread = OutLogThread(self, out_r, 'out')
                out_thread.start()
                p.wait()
                returncode = p.returncode
                
                os.close(out_r)
                os.close(err_r)
                os.close(out_w)
                os.close(err_w)
                err_thread.join()
                out_thread.join()
                        
                if returncode == 0:
                    self._status = 'Online'
                    self._send(log_message('main', 'Installation complete.'))
                    self._send(change_installation_status('installed'))
                    self._send(status_response(self._status))
                else:
                    self._status = 'Error'
                    self._send(log_message('main', 'Installation ended with error.'))
                    self._send(change_installation_status('error'))
                    self._send(status_response(self._status))
            except KeyboardInterrupt:
                logging.info('Interrupted')
                self._status = 'Error'
                self._send(log_message('main', 'Installation abort.'))
                self._send(change_installation_status('aborted'))
                self._send(status_response(self._status))
            except Exception as e:
                self._status = 'Error'
                self._send(log_message('err', str(e)))
                self._send(log_message('main', 'Installation ended with error.'))
                self._send(change_installation_status('error'))
                self._send(status_response(self._status))
                logging.error(e)
            finally:
                try:
                    os.close(out_r)
                    os.close(out_w)
                    os.close(err_r)
                    os.close(err_w)
                    err_thread.join()
                    out_thread.join()    
                except Exception as e:
                    logging.error(e)
        else:
            logging.error('Installation is already running. Abort.')
            
            
    def set_agent_name(self, agent_name):
        self._agent_name = agent_name


def configure():
    """ Reads the config-file "settings.json". 
        @returns Dict, containing:
            * server - Junta-Server address (for WebSocket connection);
            * id - agent identifier.
    """
    file_name = os.path.join('/etc/junta', 'agent.json')
    try:
        with open(file_name) as f:
            settings = json.load(f)
    except IOError:
        logging.error('Could not open config file "{}"'.format(file_name))
        exit()
        
    if 'server' not in settings:
        logging.error('"server" option must be specified')
        exit()
    
    if 'id' not in settings or settings['id'] is None:
        import uuid
        settings['id'] = str(uuid.uuid4())
        with open(file_name, 'w') as f:
            json.dump(settings, f, indent=4)                        
    return settings


def run_agent(settings=None):
    try:
        setup_logger('/var/log/junta-agent.log')
    except IOError:
        print >>sys.stderr, 'Could not setup logging to file'

    if not settings:
        settings = configure()
    
    try:
        host = settings['server']
        agent_id = settings['id']
    except Exception as e:
        logging.error('Inconsistent configuration. Error message: {0}'.format(str(e)))
        exit()
            
    try:
        ws = AgentWebSocket('ws://{host}'.format(host=host))
        ws.set_agent_name(agent_id)
        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        logging.info("Quit (Interrupted)")
        exit(1)
    except Exception as e:
        logging.error(cgitb.text(sys.exc_info()))


if __name__ == '__main__':
    run_agent()