# coding: utf-8

import time
import datetime
import logging
import pymongo
import uuid

from sharedvariables import agents_list

DB_NAME = 'testdb'
db = pymongo.Connection()[DB_NAME]


""" *** Интерфейс для работы с агентами (серверами) *** """

def agent_login(agent_id):
    """ Функция выполняется когда подключается агент.
        Если он новый, он записывается в бд и ему присваивается имя равное его id
    """
    agent = db.agents.find_one({'_id': agent_id})
    if not agent:
        name = agent_id
    else:
        name = agent['name']    
    db.agents.save({'_id': agent_id, 'name': name, 'recent_installation': '', 'status': 'Online'})
    
def get_agent_status(agent_id):
    """ Установка статуса для агента
    """
    try:
        if agent_id in agents_list:
            return db.agents.find_one({'_id': agent_id})['status']
        else:
            return 'Offline'
    except:
        return 'Offline'
    
def set_agent_status(agent_id, status):
    """ Установка статуса для агента
    """
    agent = db.agents.find_one({'_id': agent_id})
    agent['status'] = status
    db.agents.save(agent)
    
def get_agent_recent_installation(agent_id):
    """ Возвращает id последней инсталляции на сервер
    """
    return db.agents.find_one({'_id': agent_id})['recent_installation']
    
def set_agent_recent_installation(agent_id, installation_id=0):
    """ Установка id последней инсталляции на сервер
    """
    agent = db.agents.find_one({'_id': agent_id})
    agent['recent_installation'] = installation_id
    db.agents.save(agent)

def get_agent_name(agent_id):
    """ Получение имени агента по его id
    """
    agent = db.agents.find_one({'_id': agent_id})
    if agent:
        return agent['name']
    else:
        return None

def get_agents_names():
    """ Получение списка имен всех агентов
    """
    agents = db.agents.find()
    return [a['name'] for a in agents]
    
def rename_agent(agent_id, new_name):
    """ Переименнование агента
    """
    agent = db.agents.find_one({'_id': agent_id})
    agent['name'] = new_name
    db.agents.save(agent) 

def delete_agent(agent_id):
    """ Удаление агента из бд
    """
    db.agents.remove({'_id': agent_id})
    
def get_agents_list():
    """ Возвращает список словарей из id, name всех агентов
    """
    return [{'id': a['_id'],
            'name': a['name'],
            'status': get_agent_status(a['_id']),
            'log': get_recent_installation_log(a['_id'])} for a in db.agents.find()]


""" *** Интерфейс для работы с установками *** """

def get_recipes_on_server(server_id):
    """ Возвращает список id всех рецептов, когда-либо устанавливаемых на сервер
    """
    recipes = set()
    for r in db.installed.find({'server_id': server_id}):
        recipes.add(r['recipe_id'])
    return recipes

def new_installation(server_id, recipe_id, recipe_version, config, hidden_config_keys):
    """ Записывает в бд запрос на новую инсталляцию.
        Принимает: 
            server_id - id агента
            recipe_id - id рецепта
            recipe_version -  номер версии рецета
            config - словарь из параметров установки
            hidden_config_keys -список имен ключей, которые не надо записывать в бд
    """
    db_config = {}
    db_config.update(config)
    for key in hidden_config_keys:
        del db_config[key]
    installation_id = str(uuid.uuid4()).replace('-', '_')
    db.installed.save({'_id': installation_id,
                       'server_id': server_id,
                       'recipe_id': recipe_id,
                       'recipe_version': recipe_version,
                       'date': time.time(),
                       'status': 'sent',
                       'config': db_config})
    set_agent_recent_installation(server_id, installation_id)

def get_installations(server_id):
    """ Возвращает все инсталляции на сервере
        список кортежей (id рецепта, имя рецепта, список инстлляций)
    """
    recipes_on_server = get_recipes_on_server(server_id)
    installations = []
    for recipe_id in recipes_on_server:
        installations.append((recipe_id, get_recipe_name(recipe_id), get_installations_by_recipe(server_id, recipe_id)))
    return installations

def get_installations_by_recipe(server_id, recipe_id):
    """ Возвращает инсталляции определенного рецепта на определенном сервере
        (список словарей)
    """
    installations = db.installed.find({'server_id': server_id, 'recipe_id': recipe_id}).sort('date', direction=pymongo.DESCENDING)
    res = [dict(i) for i in installations]
    for r in res:
        r['id'] = str(r['_id'])
        r['date'] = str(datetime.datetime.fromtimestamp(r['date']))[:19]
        r['log'] = get_installation_log(r['id'])
    return res

def get_installations_by_time(server_id):
    """ Возвращает инсталляции определенного рецепта на определенном сервере
        (список словарей)
    """
    installations = db.installed.find({'server_id': server_id}).sort('date', direction=pymongo.DESCENDING)
    res = [dict(i) for i in installations]
    for r in res:
        r['id'] = str(r['_id'])
        r['date'] = str(datetime.datetime.fromtimestamp(r['date']))[:19]
        r['log'] = get_installation_log(r['id'])
        r['recipe_name'] = get_recipe_name(r['recipe_id'])
    return res

def set_installation_status(server_id, status):
    """ Устанавливает новый статус инсталляции.
        Для какой инсталляции, смотрит recent_installation у агента
    """
    installation_id = get_agent_recent_installation(server_id)
    if installation_id == 0:
        logging.error("error in set_installation_status")
        return
    to_change = db.installed.find_one({'_id': installation_id})
    to_change['status'] = status
    db.installed.save(to_change)
    
def save_log(server_id, msg_type, msg, n):
    """ Сохраняет сообщение в лог.
    """
    installation_id = get_agent_recent_installation(server_id)
    if installation_id == 0:
        logging.error("error in save_log")
        return
    db.log.save({'installation_id': installation_id,
                 'type': msg_type,
                 'message': msg,
                 'n': n})
    
def get_installation_log(installation_id):
    """  Получение лога по его id
    """
    def cmp_func(a, b):
        if int(a[0]) > int(b[0]):
            return 1
        elif int(b[0]) == int(a[0]):
            return 0
        else:
            return -1
    
    res = db.log.find({'installation_id': installation_id})
    log = []
    for r in res:
        log.append((r['n'], r['type'], r['message']))
    log.sort(cmp=cmp_func)
    return log
    
def get_recent_installation_log(server_id):
    """  Получение лога последней установки на сервер
    """
    installation_id = get_agent_recent_installation(server_id)
    if installation_id == 0:
        logging.error("error in get_recent_installation_log")
        return
    return get_installation_log(installation_id)

    
""" *** Интерфейс для работы с рецептами *** """

def is_recipe_exists(recipe_name):
    """ Есть ли уже рецепт с таким именем?
    """
    recipe = db.recipes.find_one({'name': recipe_name})
    return recipe is not None  

def save_recipe(recipe_name):
    """ Сохраняет новый рецепт в базу данных, генерирует id.
        Если рецепт с таким именем существует, возвращает None
    """
    recipe = db.recipes.find_one({'name': recipe_name})
    if not recipe:
        id = str(uuid.uuid4()).replace('-', '_')
        db.recipes.save({'_id': id, 'name': recipe_name})
        return id
    else:
        return None    

def get_recipe_name(recipe_id):
    """ Возвращает имя рецепта по его id.
        Если рецепт с таким id не существует, возвращает None
    """
    recipe = db.recipes.find_one({'_id': recipe_id})
    if recipe: 
        return recipe['name']
    else:
        return "Удален"
    
def get_recipe_id(recipe_name):
    """ Возвращает id рецепта по его имени.
        Если рецепт с таким именем не существует, возвращает None
    """
    recipe = db.recipes.find_one({'name': recipe_name})
    if recipe: 
        return recipe['_id']
    else:
        return None

def get_recipe_names():
    """ Возвращает список из имен всех рецептов
    """
    recipes = db.recipes.find()
    return [a['name'] for a in recipes]

def rename_recipe(recipe_id, new_name):
    """ Переименнование рецепта с заданным id.
        Если рецепт с таким именем существует, возвращает None и не переименновывает
    """
    check_recipe = db.recipes.find_one({'name': new_name})
    if check_recipe:
        return None  
    recipe = db.recipes.find_one({'_id': recipe_id})
    recipe['name'] = new_name
    db.recipes.save(recipe) 

def delete_recipe(recipe_id):
    """ Удаление рецепта по его id
    """
    db.recipes.remove({'_id': recipe_id})
    
def get_recipes_list():
    """ Возвращает список кортежей (id, name) всех рецептов
    """
    return [(x['_id'], x['name']) for x in db.recipes.find()]