# -*- coding: utf-8 -*-

try:
    from datetime import date
    import cgitb
    import json
    import glob
    import os
    import logging
    import sys
    import time
    
    from xml.etree import ElementTree
    
    from dbinterface import save_recipe, get_recipe_name, get_recipes_list, save_recipe, get_recipe_id, delete_recipe
    from settings import RECIPES_PATH
    from settings import BASE_PATH    
except Exception as e:
    logging.error(e)
    logging.error(cgitb.text(sys.exc_info()))

def get_version_number(version_file_name):
    """ Возвращает номер рецепта по имени файла
    """
    return version_file_name[1:-4]

def get_default_recipe_info(version_file_name):
    """ Возвращает дефолтную конфигурацию версии рецепта
    """
    return {"file": version_file_name, "who": "", "when": ""}

def get_recipes():
    """ Возвращает список кортежей вида (имя рецепта, id рецепта, список версий)
    """
    res = []
    try:
        for recipe_id, recipe_name in get_recipes_list():
            versions_files = [v.split('/')[-1] for v in glob.glob(os.path.join(RECIPES_PATH, recipe_id, 'v*.xml')) if os.path.isfile(v)]
            try:
                versions_desc = json.load(open(os.path.join(RECIPES_PATH, recipe_id, 'about.json'), 'r'))
                versions_names_in_json = []
                for v in versions_desc:
                    versions_names_in_json.append(v['file'])
                versions = []
                for v in versions_files:
                    if v in versions_names_in_json:
                        for i in range(0, len(versions_desc)):
                            if versions_desc[i]['file'] == v:
                                versions.append(versions_desc[i])
                                break
                    else:
                        versions.append(get_default_recipe_info(v))
            except Exception:
                versions = [get_default_recipe_info(v) for v in versions_files]
            
            for v in versions:
                v['name'] = get_version_number(v['file'])
                
            def cmp(a, b):
                if int(a['name']) > int(b['name']):
                    return 1
                elif int(a['name']) < int(b['name']):
                    return -1
                else:
                    return 0   
                
            versions.sort(reverse=True, cmp=cmp)
            res.append((recipe_name, recipe_id, versions))
    except Exception as e:
        logging.error(e)
        logging.error(cgitb.text(sys.exc_info()))
    return res

def get_last_version(recipe_id, recipes):
    """ Возвращает номер последней версии для рецепта с заданным id
    """
    try:
        for curr_name, curr_id, versions in recipes:
            if curr_id == recipe_id:
                max_version = 0
                for v in versions:
                    if int(v['name']) > max_version:
                        max_version = int(v['name'])
                return max_version
    except Exception as e:
        logging.error(e)
        logging.error(cgitb.text(sys.exc_info()))
    return None

def get_next_version(recipe_id, recipes):
    """ Возвращает номер следующей версии для рецепта с заданным id
    """
    last_version = get_last_version(recipe_id, recipes)
    if last_version is not None:
        return str(last_version + 1)
    else:
        return '1'   
    
def get_version_count(recipe_id, recipes):
    """ Возвращает количество версий рецепта с заданным id
    """
    try:
        for curr_name, curr_id, versions in recipes:
            if curr_id == recipe_id:
                return len(versions)
    except Exception as e:
        logging.error(e)
        logging.error(cgitb.text(sys.exc_info()))
    return 0    

def get_default_config(recipe_id, recipe_version):
    """ Принимает id рецепта и номер версии
        Возвращает кортеж (имя параметра, дефолтное значение, тип)
    """
    try:
        tree = ElementTree.parse(get_recipe_file_name(recipe_id, recipe_version))
        conf = []
        for arg in tree.findall('./Args/Arg'):
            name = arg.text
            default = arg.get('default', '')
            if arg.get('type') == 'password':
                conf.append((name, default, 'password'))
            else:
                conf.append((name, default, 'text'))
        return conf
    except Exception as e:
        logging.error(e)
        logging.error(cgitb.text(sys.exc_info()))
            
def not_valid_error_message(recipe):
    """ Проверяет, валиден ли документ. принимет текст документа.
        Если все нормально, возвращает None, иначе сообщение об ошибке
    """
    try:
        ElementTree.fromstring(recipe.encode('utf-8'))
    except Exception as e:
        return e
    return None

def get_recipe_file_name(recipe_id, recipe_version):
    """ Возвращает путь до файла с рецептом, принимает id рецепта и версию
    """
    return os.path.join(RECIPES_PATH, recipe_id, 'v' + str(recipe_version) + '.xml')

def create_recipe_path_if_not_exists(recipe_id):
    """ Создает папку для рецепта, если она не существует
    """
    if not os.path.exists(os.path.join(RECIPES_PATH, recipe_id)):
        os.mkdir(os.path.join(RECIPES_PATH, recipe_id))
        
def get_date():
    """ Возвращает сегодняшнюю дату
    """
    d = date.today()
    return str(d.day) + '.' + str(d.month) + '.' + str(d.year) 
        
def recipe_save(recipe_name, version, content, username):
    """ Сохраняет рецепт на файловую систему, 
        принимает название рецепта, версию, содержимое и юзернэйм
    """
    try:
        recipe_id = save_recipe(recipe_name)
        if recipe_id is None: # рецепт с таким названием уже есть
            recipe_id = get_recipe_id(recipe_name)
        
        open(get_recipe_file_name(recipe_id, version), 'w').write(content.strip().encode('utf-8'))
        
        try:
            versions = json.load(open(os.path.join(RECIPES_PATH, recipe_id, 'about.json'), 'r'))
            version_names = [get_version_number(v['file']) for v in versions]
            if version in version_names:
                return 'Такая версия уже существует'
        except Exception as e:
            versions = []
            
        versions.append({'file': 'v' + str(version) + '.xml',
                       'who': username,
                       'when': get_date()})
        
        json.dump(versions, open(os.path.join(RECIPES_PATH, recipe_id, 'about.json'), 'w'))
        return None #ok
    except Exception as e:
        logging.error(e)
        logging.error(cgitb.text(sys.exc_info()))
        return 'Неизвестная ошибка. Смотрите лог.'
    
    
def get_recipe_content(recipe_id, version):
    """ Возвращает содержимое рецепта с заданным id
    """
    try:
        return open(get_recipe_file_name(recipe_id, version), 'r').read().decode('utf-8')
    except:
        return ""

def get_recipe_template():
    """ Возвращает содержимое шаблона, если он существует
    """
    try:
        return open(os.path.join(RECIPES_PATH, "recipe_template.xml"), 'r').read().decode('utf-8')
    except:
        return ""

def get_setup_tools():
    """ Возвращает содержимое файла setup_tools
    """
    return open(os.path.join(BASE_PATH, 'setup_tools.py'), 'r').read().decode('utf-8')

def delete_recipe_version(recipe_id, recipe_version):
    """ Удаление версии рецепта
    """
    versions = json.load(open(os.path.join(RECIPES_PATH, recipe_id, 'about.json'), 'r'))
    file_name = 'v' + str(recipe_version) + '.xml'
    new_versions = [x for x in versions if x['file'] != file_name]
    json.dump(new_versions, open(os.path.join(RECIPES_PATH, recipe_id, 'about.json'), 'w'))
    os.remove(os.path.join(RECIPES_PATH, recipe_id, file_name))
    
def recipe_delete(recipe_id):
    """ Удаление рецепта полностью
    """
    import shutil
    shutil.rmtree(os.path.join(RECIPES_PATH, recipe_id))
    delete_recipe(recipe_id)
    
def preprocessor(content, direction=True):
    """ Препроцессор для рецептов. 
        Если direction == True, то это из текста в xml (при вводе рецепта),
        Если direction == False, то это из xml в текст (при выводе рецепта)
    """
    dictionary_to_replace = {'&': '&amp;',
                             '<': '&gt;',
                             '>': '&lt;',
                             '\'': '&apos;',
                             '\"': '&quot;'}
    new_content = content
    if direction:
        for key, val in dictionary_to_replace.items():
            new_content = new_content.replace(key, val);
    else: 
        for val, key in dictionary_to_replace.items():
            new_content = new_content.replace(key, val);
    return new_content